
# THIS FILE IS AUTO-GENERATED!!!
# DO NOT MODIFY THIS FILE.
# CHANGES TO THE CODE SHOULD BE MADE IN 'opcodes_gen.py'.

cimport cython
from libc.stdint cimport uint8_t, uint16_t, uint32_t

from pyboy.logging.logging cimport Logger

from . cimport cpu


cdef Logger logger

cdef uint16_t FLAGC, FLAGH, FLAGN, FLAGZ
cdef uint8_t[512] OPCODE_LENGTHS
@cython.locals(v=cython.int, a=cython.int, b=cython.int, pc=cython.ushort)
cdef int execute_opcode(cpu.CPU, uint16_t) noexcept nogil

cdef uint8_t no_opcode(cpu.CPU) noexcept nogil
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t NOP_00(cpu.CPU) noexcept nogil # 00 NOP
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_01(cpu.CPU, int v) noexcept nogil # 01 LD BC,d16
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_02(cpu.CPU) noexcept nogil # 02 LD (BC),A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t INC_03(cpu.CPU) noexcept nogil # 03 INC BC
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t INC_04(cpu.CPU) noexcept nogil # 04 INC B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t DEC_05(cpu.CPU) noexcept nogil # 05 DEC B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_06(cpu.CPU, int v) noexcept nogil # 06 LD B,d8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RLCA_07(cpu.CPU) noexcept nogil # 07 RLCA
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_08(cpu.CPU, int v) noexcept nogil # 08 LD (a16),SP
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADD_09(cpu.CPU) noexcept nogil # 09 ADD HL,BC
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_0A(cpu.CPU) noexcept nogil # 0A LD A,(BC)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t DEC_0B(cpu.CPU) noexcept nogil # 0B DEC BC
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t INC_0C(cpu.CPU) noexcept nogil # 0C INC C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t DEC_0D(cpu.CPU) noexcept nogil # 0D DEC C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_0E(cpu.CPU, int v) noexcept nogil # 0E LD C,d8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RRCA_0F(cpu.CPU) noexcept nogil # 0F RRCA
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t STOP_10(cpu.CPU, int v) noexcept nogil # 10 STOP 0
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_11(cpu.CPU, int v) noexcept nogil # 11 LD DE,d16
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_12(cpu.CPU) noexcept nogil # 12 LD (DE),A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t INC_13(cpu.CPU) noexcept nogil # 13 INC DE
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t INC_14(cpu.CPU) noexcept nogil # 14 INC D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t DEC_15(cpu.CPU) noexcept nogil # 15 DEC D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_16(cpu.CPU, int v) noexcept nogil # 16 LD D,d8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RLA_17(cpu.CPU) noexcept nogil # 17 RLA
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t JR_18(cpu.CPU, int v) noexcept nogil # 18 JR r8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADD_19(cpu.CPU) noexcept nogil # 19 ADD HL,DE
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_1A(cpu.CPU) noexcept nogil # 1A LD A,(DE)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t DEC_1B(cpu.CPU) noexcept nogil # 1B DEC DE
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t INC_1C(cpu.CPU) noexcept nogil # 1C INC E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t DEC_1D(cpu.CPU) noexcept nogil # 1D DEC E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_1E(cpu.CPU, int v) noexcept nogil # 1E LD E,d8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RRA_1F(cpu.CPU) noexcept nogil # 1F RRA
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t JR_20(cpu.CPU, int v) noexcept nogil # 20 JR NZ,r8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_21(cpu.CPU, int v) noexcept nogil # 21 LD HL,d16
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_22(cpu.CPU) noexcept nogil # 22 LD (HL+),A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t INC_23(cpu.CPU) noexcept nogil # 23 INC HL
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t INC_24(cpu.CPU) noexcept nogil # 24 INC H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t DEC_25(cpu.CPU) noexcept nogil # 25 DEC H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_26(cpu.CPU, int v) noexcept nogil # 26 LD H,d8
@cython.locals(v=int, flag=uint8_t, t=int, corr=ushort)
cdef uint8_t DAA_27(cpu.CPU) noexcept nogil # 27 DAA
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t JR_28(cpu.CPU, int v) noexcept nogil # 28 JR Z,r8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADD_29(cpu.CPU) noexcept nogil # 29 ADD HL,HL
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_2A(cpu.CPU) noexcept nogil # 2A LD A,(HL+)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t DEC_2B(cpu.CPU) noexcept nogil # 2B DEC HL
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t INC_2C(cpu.CPU) noexcept nogil # 2C INC L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t DEC_2D(cpu.CPU) noexcept nogil # 2D DEC L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_2E(cpu.CPU, int v) noexcept nogil # 2E LD L,d8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t CPL_2F(cpu.CPU) noexcept nogil # 2F CPL
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t JR_30(cpu.CPU, int v) noexcept nogil # 30 JR NC,r8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_31(cpu.CPU, int v) noexcept nogil # 31 LD SP,d16
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_32(cpu.CPU) noexcept nogil # 32 LD (HL-),A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t INC_33(cpu.CPU) noexcept nogil # 33 INC SP
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t INC_34(cpu.CPU) noexcept nogil # 34 INC (HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t DEC_35(cpu.CPU) noexcept nogil # 35 DEC (HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_36(cpu.CPU, int v) noexcept nogil # 36 LD (HL),d8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SCF_37(cpu.CPU) noexcept nogil # 37 SCF
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t JR_38(cpu.CPU, int v) noexcept nogil # 38 JR C,r8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADD_39(cpu.CPU) noexcept nogil # 39 ADD HL,SP
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_3A(cpu.CPU) noexcept nogil # 3A LD A,(HL-)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t DEC_3B(cpu.CPU) noexcept nogil # 3B DEC SP
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t INC_3C(cpu.CPU) noexcept nogil # 3C INC A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t DEC_3D(cpu.CPU) noexcept nogil # 3D DEC A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_3E(cpu.CPU, int v) noexcept nogil # 3E LD A,d8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t CCF_3F(cpu.CPU) noexcept nogil # 3F CCF
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_40(cpu.CPU) noexcept nogil # 40 LD B,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_41(cpu.CPU) noexcept nogil # 41 LD B,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_42(cpu.CPU) noexcept nogil # 42 LD B,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_43(cpu.CPU) noexcept nogil # 43 LD B,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_44(cpu.CPU) noexcept nogil # 44 LD B,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_45(cpu.CPU) noexcept nogil # 45 LD B,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_46(cpu.CPU) noexcept nogil # 46 LD B,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_47(cpu.CPU) noexcept nogil # 47 LD B,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_48(cpu.CPU) noexcept nogil # 48 LD C,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_49(cpu.CPU) noexcept nogil # 49 LD C,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_4A(cpu.CPU) noexcept nogil # 4A LD C,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_4B(cpu.CPU) noexcept nogil # 4B LD C,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_4C(cpu.CPU) noexcept nogil # 4C LD C,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_4D(cpu.CPU) noexcept nogil # 4D LD C,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_4E(cpu.CPU) noexcept nogil # 4E LD C,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_4F(cpu.CPU) noexcept nogil # 4F LD C,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_50(cpu.CPU) noexcept nogil # 50 LD D,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_51(cpu.CPU) noexcept nogil # 51 LD D,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_52(cpu.CPU) noexcept nogil # 52 LD D,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_53(cpu.CPU) noexcept nogil # 53 LD D,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_54(cpu.CPU) noexcept nogil # 54 LD D,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_55(cpu.CPU) noexcept nogil # 55 LD D,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_56(cpu.CPU) noexcept nogil # 56 LD D,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_57(cpu.CPU) noexcept nogil # 57 LD D,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_58(cpu.CPU) noexcept nogil # 58 LD E,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_59(cpu.CPU) noexcept nogil # 59 LD E,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_5A(cpu.CPU) noexcept nogil # 5A LD E,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_5B(cpu.CPU) noexcept nogil # 5B LD E,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_5C(cpu.CPU) noexcept nogil # 5C LD E,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_5D(cpu.CPU) noexcept nogil # 5D LD E,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_5E(cpu.CPU) noexcept nogil # 5E LD E,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_5F(cpu.CPU) noexcept nogil # 5F LD E,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_60(cpu.CPU) noexcept nogil # 60 LD H,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_61(cpu.CPU) noexcept nogil # 61 LD H,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_62(cpu.CPU) noexcept nogil # 62 LD H,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_63(cpu.CPU) noexcept nogil # 63 LD H,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_64(cpu.CPU) noexcept nogil # 64 LD H,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_65(cpu.CPU) noexcept nogil # 65 LD H,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_66(cpu.CPU) noexcept nogil # 66 LD H,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_67(cpu.CPU) noexcept nogil # 67 LD H,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_68(cpu.CPU) noexcept nogil # 68 LD L,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_69(cpu.CPU) noexcept nogil # 69 LD L,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_6A(cpu.CPU) noexcept nogil # 6A LD L,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_6B(cpu.CPU) noexcept nogil # 6B LD L,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_6C(cpu.CPU) noexcept nogil # 6C LD L,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_6D(cpu.CPU) noexcept nogil # 6D LD L,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_6E(cpu.CPU) noexcept nogil # 6E LD L,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_6F(cpu.CPU) noexcept nogil # 6F LD L,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_70(cpu.CPU) noexcept nogil # 70 LD (HL),B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_71(cpu.CPU) noexcept nogil # 71 LD (HL),C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_72(cpu.CPU) noexcept nogil # 72 LD (HL),D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_73(cpu.CPU) noexcept nogil # 73 LD (HL),E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_74(cpu.CPU) noexcept nogil # 74 LD (HL),H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_75(cpu.CPU) noexcept nogil # 75 LD (HL),L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t HALT_76(cpu.CPU) noexcept nogil # 76 HALT
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_77(cpu.CPU) noexcept nogil # 77 LD (HL),A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_78(cpu.CPU) noexcept nogil # 78 LD A,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_79(cpu.CPU) noexcept nogil # 79 LD A,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_7A(cpu.CPU) noexcept nogil # 7A LD A,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_7B(cpu.CPU) noexcept nogil # 7B LD A,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_7C(cpu.CPU) noexcept nogil # 7C LD A,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_7D(cpu.CPU) noexcept nogil # 7D LD A,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_7E(cpu.CPU) noexcept nogil # 7E LD A,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_7F(cpu.CPU) noexcept nogil # 7F LD A,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADD_80(cpu.CPU) noexcept nogil # 80 ADD A,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADD_81(cpu.CPU) noexcept nogil # 81 ADD A,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADD_82(cpu.CPU) noexcept nogil # 82 ADD A,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADD_83(cpu.CPU) noexcept nogil # 83 ADD A,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADD_84(cpu.CPU) noexcept nogil # 84 ADD A,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADD_85(cpu.CPU) noexcept nogil # 85 ADD A,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADD_86(cpu.CPU) noexcept nogil # 86 ADD A,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADD_87(cpu.CPU) noexcept nogil # 87 ADD A,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADC_88(cpu.CPU) noexcept nogil # 88 ADC A,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADC_89(cpu.CPU) noexcept nogil # 89 ADC A,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADC_8A(cpu.CPU) noexcept nogil # 8A ADC A,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADC_8B(cpu.CPU) noexcept nogil # 8B ADC A,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADC_8C(cpu.CPU) noexcept nogil # 8C ADC A,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADC_8D(cpu.CPU) noexcept nogil # 8D ADC A,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADC_8E(cpu.CPU) noexcept nogil # 8E ADC A,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADC_8F(cpu.CPU) noexcept nogil # 8F ADC A,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SUB_90(cpu.CPU) noexcept nogil # 90 SUB B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SUB_91(cpu.CPU) noexcept nogil # 91 SUB C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SUB_92(cpu.CPU) noexcept nogil # 92 SUB D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SUB_93(cpu.CPU) noexcept nogil # 93 SUB E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SUB_94(cpu.CPU) noexcept nogil # 94 SUB H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SUB_95(cpu.CPU) noexcept nogil # 95 SUB L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SUB_96(cpu.CPU) noexcept nogil # 96 SUB (HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SUB_97(cpu.CPU) noexcept nogil # 97 SUB A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SBC_98(cpu.CPU) noexcept nogil # 98 SBC A,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SBC_99(cpu.CPU) noexcept nogil # 99 SBC A,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SBC_9A(cpu.CPU) noexcept nogil # 9A SBC A,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SBC_9B(cpu.CPU) noexcept nogil # 9B SBC A,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SBC_9C(cpu.CPU) noexcept nogil # 9C SBC A,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SBC_9D(cpu.CPU) noexcept nogil # 9D SBC A,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SBC_9E(cpu.CPU) noexcept nogil # 9E SBC A,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SBC_9F(cpu.CPU) noexcept nogil # 9F SBC A,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t AND_A0(cpu.CPU) noexcept nogil # A0 AND B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t AND_A1(cpu.CPU) noexcept nogil # A1 AND C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t AND_A2(cpu.CPU) noexcept nogil # A2 AND D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t AND_A3(cpu.CPU) noexcept nogil # A3 AND E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t AND_A4(cpu.CPU) noexcept nogil # A4 AND H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t AND_A5(cpu.CPU) noexcept nogil # A5 AND L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t AND_A6(cpu.CPU) noexcept nogil # A6 AND (HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t AND_A7(cpu.CPU) noexcept nogil # A7 AND A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t XOR_A8(cpu.CPU) noexcept nogil # A8 XOR B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t XOR_A9(cpu.CPU) noexcept nogil # A9 XOR C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t XOR_AA(cpu.CPU) noexcept nogil # AA XOR D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t XOR_AB(cpu.CPU) noexcept nogil # AB XOR E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t XOR_AC(cpu.CPU) noexcept nogil # AC XOR H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t XOR_AD(cpu.CPU) noexcept nogil # AD XOR L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t XOR_AE(cpu.CPU) noexcept nogil # AE XOR (HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t XOR_AF(cpu.CPU) noexcept nogil # AF XOR A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t OR_B0(cpu.CPU) noexcept nogil # B0 OR B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t OR_B1(cpu.CPU) noexcept nogil # B1 OR C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t OR_B2(cpu.CPU) noexcept nogil # B2 OR D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t OR_B3(cpu.CPU) noexcept nogil # B3 OR E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t OR_B4(cpu.CPU) noexcept nogil # B4 OR H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t OR_B5(cpu.CPU) noexcept nogil # B5 OR L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t OR_B6(cpu.CPU) noexcept nogil # B6 OR (HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t OR_B7(cpu.CPU) noexcept nogil # B7 OR A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t CP_B8(cpu.CPU) noexcept nogil # B8 CP B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t CP_B9(cpu.CPU) noexcept nogil # B9 CP C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t CP_BA(cpu.CPU) noexcept nogil # BA CP D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t CP_BB(cpu.CPU) noexcept nogil # BB CP E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t CP_BC(cpu.CPU) noexcept nogil # BC CP H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t CP_BD(cpu.CPU) noexcept nogil # BD CP L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t CP_BE(cpu.CPU) noexcept nogil # BE CP (HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t CP_BF(cpu.CPU) noexcept nogil # BF CP A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RET_C0(cpu.CPU) noexcept nogil # C0 RET NZ
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t POP_C1(cpu.CPU) noexcept nogil # C1 POP BC
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t JP_C2(cpu.CPU, int v) noexcept nogil # C2 JP NZ,a16
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t JP_C3(cpu.CPU, int v) noexcept nogil # C3 JP a16
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t CALL_C4(cpu.CPU, int v) noexcept nogil # C4 CALL NZ,a16
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t PUSH_C5(cpu.CPU) noexcept nogil # C5 PUSH BC
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADD_C6(cpu.CPU, int v) noexcept nogil # C6 ADD A,d8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RST_C7(cpu.CPU) noexcept nogil # C7 RST 00H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RET_C8(cpu.CPU) noexcept nogil # C8 RET Z
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RET_C9(cpu.CPU) noexcept nogil # C9 RET
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t JP_CA(cpu.CPU, int v) noexcept nogil # CA JP Z,a16
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t PREFIX_CB(cpu.CPU) noexcept nogil # CB PREFIX CB
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t CALL_CC(cpu.CPU, int v) noexcept nogil # CC CALL Z,a16
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t CALL_CD(cpu.CPU, int v) noexcept nogil # CD CALL a16
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADC_CE(cpu.CPU, int v) noexcept nogil # CE ADC A,d8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RST_CF(cpu.CPU) noexcept nogil # CF RST 08H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RET_D0(cpu.CPU) noexcept nogil # D0 RET NC
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t POP_D1(cpu.CPU) noexcept nogil # D1 POP DE
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t JP_D2(cpu.CPU, int v) noexcept nogil # D2 JP NC,a16
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t CALL_D4(cpu.CPU, int v) noexcept nogil # D4 CALL NC,a16
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t PUSH_D5(cpu.CPU) noexcept nogil # D5 PUSH DE
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SUB_D6(cpu.CPU, int v) noexcept nogil # D6 SUB d8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RST_D7(cpu.CPU) noexcept nogil # D7 RST 10H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RET_D8(cpu.CPU) noexcept nogil # D8 RET C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RETI_D9(cpu.CPU) noexcept nogil # D9 RETI
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t JP_DA(cpu.CPU, int v) noexcept nogil # DA JP C,a16
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t CALL_DC(cpu.CPU, int v) noexcept nogil # DC CALL C,a16
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SBC_DE(cpu.CPU, int v) noexcept nogil # DE SBC A,d8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RST_DF(cpu.CPU) noexcept nogil # DF RST 18H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LDH_E0(cpu.CPU, int v) noexcept nogil # E0 LDH (a8),A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t POP_E1(cpu.CPU) noexcept nogil # E1 POP HL
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_E2(cpu.CPU) noexcept nogil # E2 LD (C),A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t PUSH_E5(cpu.CPU) noexcept nogil # E5 PUSH HL
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t AND_E6(cpu.CPU, int v) noexcept nogil # E6 AND d8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RST_E7(cpu.CPU) noexcept nogil # E7 RST 20H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t ADD_E8(cpu.CPU, int v) noexcept nogil # E8 ADD SP,r8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t JP_E9(cpu.CPU) noexcept nogil # E9 JP (HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_EA(cpu.CPU, int v) noexcept nogil # EA LD (a16),A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t XOR_EE(cpu.CPU, int v) noexcept nogil # EE XOR d8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RST_EF(cpu.CPU) noexcept nogil # EF RST 28H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LDH_F0(cpu.CPU, int v) noexcept nogil # F0 LDH A,(a8)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t POP_F1(cpu.CPU) noexcept nogil # F1 POP AF
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_F2(cpu.CPU) noexcept nogil # F2 LD A,(C)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t DI_F3(cpu.CPU) noexcept nogil # F3 DI
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t PUSH_F5(cpu.CPU) noexcept nogil # F5 PUSH AF
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t OR_F6(cpu.CPU, int v) noexcept nogil # F6 OR d8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RST_F7(cpu.CPU) noexcept nogil # F7 RST 30H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_F8(cpu.CPU, int v) noexcept nogil # F8 LD HL,SP+r8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_F9(cpu.CPU) noexcept nogil # F9 LD SP,HL
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t LD_FA(cpu.CPU, int v) noexcept nogil # FA LD A,(a16)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t EI_FB(cpu.CPU) noexcept nogil # FB EI
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t CP_FE(cpu.CPU, int v) noexcept nogil # FE CP d8
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RST_FF(cpu.CPU) noexcept nogil # FF RST 38H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RLC_100(cpu.CPU) noexcept nogil # 100 RLC B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RLC_101(cpu.CPU) noexcept nogil # 101 RLC C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RLC_102(cpu.CPU) noexcept nogil # 102 RLC D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RLC_103(cpu.CPU) noexcept nogil # 103 RLC E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RLC_104(cpu.CPU) noexcept nogil # 104 RLC H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RLC_105(cpu.CPU) noexcept nogil # 105 RLC L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RLC_106(cpu.CPU) noexcept nogil # 106 RLC (HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RLC_107(cpu.CPU) noexcept nogil # 107 RLC A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RRC_108(cpu.CPU) noexcept nogil # 108 RRC B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RRC_109(cpu.CPU) noexcept nogil # 109 RRC C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RRC_10A(cpu.CPU) noexcept nogil # 10A RRC D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RRC_10B(cpu.CPU) noexcept nogil # 10B RRC E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RRC_10C(cpu.CPU) noexcept nogil # 10C RRC H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RRC_10D(cpu.CPU) noexcept nogil # 10D RRC L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RRC_10E(cpu.CPU) noexcept nogil # 10E RRC (HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RRC_10F(cpu.CPU) noexcept nogil # 10F RRC A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RL_110(cpu.CPU) noexcept nogil # 110 RL B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RL_111(cpu.CPU) noexcept nogil # 111 RL C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RL_112(cpu.CPU) noexcept nogil # 112 RL D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RL_113(cpu.CPU) noexcept nogil # 113 RL E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RL_114(cpu.CPU) noexcept nogil # 114 RL H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RL_115(cpu.CPU) noexcept nogil # 115 RL L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RL_116(cpu.CPU) noexcept nogil # 116 RL (HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RL_117(cpu.CPU) noexcept nogil # 117 RL A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RR_118(cpu.CPU) noexcept nogil # 118 RR B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RR_119(cpu.CPU) noexcept nogil # 119 RR C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RR_11A(cpu.CPU) noexcept nogil # 11A RR D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RR_11B(cpu.CPU) noexcept nogil # 11B RR E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RR_11C(cpu.CPU) noexcept nogil # 11C RR H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RR_11D(cpu.CPU) noexcept nogil # 11D RR L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RR_11E(cpu.CPU) noexcept nogil # 11E RR (HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RR_11F(cpu.CPU) noexcept nogil # 11F RR A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SLA_120(cpu.CPU) noexcept nogil # 120 SLA B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SLA_121(cpu.CPU) noexcept nogil # 121 SLA C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SLA_122(cpu.CPU) noexcept nogil # 122 SLA D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SLA_123(cpu.CPU) noexcept nogil # 123 SLA E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SLA_124(cpu.CPU) noexcept nogil # 124 SLA H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SLA_125(cpu.CPU) noexcept nogil # 125 SLA L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SLA_126(cpu.CPU) noexcept nogil # 126 SLA (HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SLA_127(cpu.CPU) noexcept nogil # 127 SLA A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SRA_128(cpu.CPU) noexcept nogil # 128 SRA B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SRA_129(cpu.CPU) noexcept nogil # 129 SRA C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SRA_12A(cpu.CPU) noexcept nogil # 12A SRA D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SRA_12B(cpu.CPU) noexcept nogil # 12B SRA E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SRA_12C(cpu.CPU) noexcept nogil # 12C SRA H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SRA_12D(cpu.CPU) noexcept nogil # 12D SRA L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SRA_12E(cpu.CPU) noexcept nogil # 12E SRA (HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SRA_12F(cpu.CPU) noexcept nogil # 12F SRA A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SWAP_130(cpu.CPU) noexcept nogil # 130 SWAP B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SWAP_131(cpu.CPU) noexcept nogil # 131 SWAP C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SWAP_132(cpu.CPU) noexcept nogil # 132 SWAP D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SWAP_133(cpu.CPU) noexcept nogil # 133 SWAP E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SWAP_134(cpu.CPU) noexcept nogil # 134 SWAP H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SWAP_135(cpu.CPU) noexcept nogil # 135 SWAP L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SWAP_136(cpu.CPU) noexcept nogil # 136 SWAP (HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SWAP_137(cpu.CPU) noexcept nogil # 137 SWAP A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SRL_138(cpu.CPU) noexcept nogil # 138 SRL B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SRL_139(cpu.CPU) noexcept nogil # 139 SRL C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SRL_13A(cpu.CPU) noexcept nogil # 13A SRL D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SRL_13B(cpu.CPU) noexcept nogil # 13B SRL E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SRL_13C(cpu.CPU) noexcept nogil # 13C SRL H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SRL_13D(cpu.CPU) noexcept nogil # 13D SRL L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SRL_13E(cpu.CPU) noexcept nogil # 13E SRL (HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SRL_13F(cpu.CPU) noexcept nogil # 13F SRL A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_140(cpu.CPU) noexcept nogil # 140 BIT 0,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_141(cpu.CPU) noexcept nogil # 141 BIT 0,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_142(cpu.CPU) noexcept nogil # 142 BIT 0,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_143(cpu.CPU) noexcept nogil # 143 BIT 0,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_144(cpu.CPU) noexcept nogil # 144 BIT 0,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_145(cpu.CPU) noexcept nogil # 145 BIT 0,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_146(cpu.CPU) noexcept nogil # 146 BIT 0,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_147(cpu.CPU) noexcept nogil # 147 BIT 0,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_148(cpu.CPU) noexcept nogil # 148 BIT 1,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_149(cpu.CPU) noexcept nogil # 149 BIT 1,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_14A(cpu.CPU) noexcept nogil # 14A BIT 1,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_14B(cpu.CPU) noexcept nogil # 14B BIT 1,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_14C(cpu.CPU) noexcept nogil # 14C BIT 1,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_14D(cpu.CPU) noexcept nogil # 14D BIT 1,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_14E(cpu.CPU) noexcept nogil # 14E BIT 1,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_14F(cpu.CPU) noexcept nogil # 14F BIT 1,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_150(cpu.CPU) noexcept nogil # 150 BIT 2,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_151(cpu.CPU) noexcept nogil # 151 BIT 2,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_152(cpu.CPU) noexcept nogil # 152 BIT 2,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_153(cpu.CPU) noexcept nogil # 153 BIT 2,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_154(cpu.CPU) noexcept nogil # 154 BIT 2,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_155(cpu.CPU) noexcept nogil # 155 BIT 2,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_156(cpu.CPU) noexcept nogil # 156 BIT 2,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_157(cpu.CPU) noexcept nogil # 157 BIT 2,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_158(cpu.CPU) noexcept nogil # 158 BIT 3,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_159(cpu.CPU) noexcept nogil # 159 BIT 3,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_15A(cpu.CPU) noexcept nogil # 15A BIT 3,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_15B(cpu.CPU) noexcept nogil # 15B BIT 3,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_15C(cpu.CPU) noexcept nogil # 15C BIT 3,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_15D(cpu.CPU) noexcept nogil # 15D BIT 3,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_15E(cpu.CPU) noexcept nogil # 15E BIT 3,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_15F(cpu.CPU) noexcept nogil # 15F BIT 3,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_160(cpu.CPU) noexcept nogil # 160 BIT 4,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_161(cpu.CPU) noexcept nogil # 161 BIT 4,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_162(cpu.CPU) noexcept nogil # 162 BIT 4,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_163(cpu.CPU) noexcept nogil # 163 BIT 4,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_164(cpu.CPU) noexcept nogil # 164 BIT 4,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_165(cpu.CPU) noexcept nogil # 165 BIT 4,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_166(cpu.CPU) noexcept nogil # 166 BIT 4,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_167(cpu.CPU) noexcept nogil # 167 BIT 4,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_168(cpu.CPU) noexcept nogil # 168 BIT 5,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_169(cpu.CPU) noexcept nogil # 169 BIT 5,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_16A(cpu.CPU) noexcept nogil # 16A BIT 5,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_16B(cpu.CPU) noexcept nogil # 16B BIT 5,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_16C(cpu.CPU) noexcept nogil # 16C BIT 5,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_16D(cpu.CPU) noexcept nogil # 16D BIT 5,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_16E(cpu.CPU) noexcept nogil # 16E BIT 5,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_16F(cpu.CPU) noexcept nogil # 16F BIT 5,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_170(cpu.CPU) noexcept nogil # 170 BIT 6,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_171(cpu.CPU) noexcept nogil # 171 BIT 6,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_172(cpu.CPU) noexcept nogil # 172 BIT 6,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_173(cpu.CPU) noexcept nogil # 173 BIT 6,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_174(cpu.CPU) noexcept nogil # 174 BIT 6,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_175(cpu.CPU) noexcept nogil # 175 BIT 6,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_176(cpu.CPU) noexcept nogil # 176 BIT 6,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_177(cpu.CPU) noexcept nogil # 177 BIT 6,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_178(cpu.CPU) noexcept nogil # 178 BIT 7,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_179(cpu.CPU) noexcept nogil # 179 BIT 7,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_17A(cpu.CPU) noexcept nogil # 17A BIT 7,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_17B(cpu.CPU) noexcept nogil # 17B BIT 7,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_17C(cpu.CPU) noexcept nogil # 17C BIT 7,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_17D(cpu.CPU) noexcept nogil # 17D BIT 7,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_17E(cpu.CPU) noexcept nogil # 17E BIT 7,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t BIT_17F(cpu.CPU) noexcept nogil # 17F BIT 7,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_180(cpu.CPU) noexcept nogil # 180 RES 0,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_181(cpu.CPU) noexcept nogil # 181 RES 0,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_182(cpu.CPU) noexcept nogil # 182 RES 0,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_183(cpu.CPU) noexcept nogil # 183 RES 0,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_184(cpu.CPU) noexcept nogil # 184 RES 0,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_185(cpu.CPU) noexcept nogil # 185 RES 0,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_186(cpu.CPU) noexcept nogil # 186 RES 0,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_187(cpu.CPU) noexcept nogil # 187 RES 0,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_188(cpu.CPU) noexcept nogil # 188 RES 1,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_189(cpu.CPU) noexcept nogil # 189 RES 1,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_18A(cpu.CPU) noexcept nogil # 18A RES 1,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_18B(cpu.CPU) noexcept nogil # 18B RES 1,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_18C(cpu.CPU) noexcept nogil # 18C RES 1,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_18D(cpu.CPU) noexcept nogil # 18D RES 1,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_18E(cpu.CPU) noexcept nogil # 18E RES 1,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_18F(cpu.CPU) noexcept nogil # 18F RES 1,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_190(cpu.CPU) noexcept nogil # 190 RES 2,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_191(cpu.CPU) noexcept nogil # 191 RES 2,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_192(cpu.CPU) noexcept nogil # 192 RES 2,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_193(cpu.CPU) noexcept nogil # 193 RES 2,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_194(cpu.CPU) noexcept nogil # 194 RES 2,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_195(cpu.CPU) noexcept nogil # 195 RES 2,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_196(cpu.CPU) noexcept nogil # 196 RES 2,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_197(cpu.CPU) noexcept nogil # 197 RES 2,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_198(cpu.CPU) noexcept nogil # 198 RES 3,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_199(cpu.CPU) noexcept nogil # 199 RES 3,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_19A(cpu.CPU) noexcept nogil # 19A RES 3,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_19B(cpu.CPU) noexcept nogil # 19B RES 3,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_19C(cpu.CPU) noexcept nogil # 19C RES 3,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_19D(cpu.CPU) noexcept nogil # 19D RES 3,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_19E(cpu.CPU) noexcept nogil # 19E RES 3,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_19F(cpu.CPU) noexcept nogil # 19F RES 3,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1A0(cpu.CPU) noexcept nogil # 1A0 RES 4,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1A1(cpu.CPU) noexcept nogil # 1A1 RES 4,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1A2(cpu.CPU) noexcept nogil # 1A2 RES 4,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1A3(cpu.CPU) noexcept nogil # 1A3 RES 4,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1A4(cpu.CPU) noexcept nogil # 1A4 RES 4,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1A5(cpu.CPU) noexcept nogil # 1A5 RES 4,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1A6(cpu.CPU) noexcept nogil # 1A6 RES 4,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1A7(cpu.CPU) noexcept nogil # 1A7 RES 4,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1A8(cpu.CPU) noexcept nogil # 1A8 RES 5,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1A9(cpu.CPU) noexcept nogil # 1A9 RES 5,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1AA(cpu.CPU) noexcept nogil # 1AA RES 5,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1AB(cpu.CPU) noexcept nogil # 1AB RES 5,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1AC(cpu.CPU) noexcept nogil # 1AC RES 5,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1AD(cpu.CPU) noexcept nogil # 1AD RES 5,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1AE(cpu.CPU) noexcept nogil # 1AE RES 5,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1AF(cpu.CPU) noexcept nogil # 1AF RES 5,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1B0(cpu.CPU) noexcept nogil # 1B0 RES 6,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1B1(cpu.CPU) noexcept nogil # 1B1 RES 6,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1B2(cpu.CPU) noexcept nogil # 1B2 RES 6,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1B3(cpu.CPU) noexcept nogil # 1B3 RES 6,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1B4(cpu.CPU) noexcept nogil # 1B4 RES 6,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1B5(cpu.CPU) noexcept nogil # 1B5 RES 6,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1B6(cpu.CPU) noexcept nogil # 1B6 RES 6,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1B7(cpu.CPU) noexcept nogil # 1B7 RES 6,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1B8(cpu.CPU) noexcept nogil # 1B8 RES 7,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1B9(cpu.CPU) noexcept nogil # 1B9 RES 7,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1BA(cpu.CPU) noexcept nogil # 1BA RES 7,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1BB(cpu.CPU) noexcept nogil # 1BB RES 7,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1BC(cpu.CPU) noexcept nogil # 1BC RES 7,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1BD(cpu.CPU) noexcept nogil # 1BD RES 7,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1BE(cpu.CPU) noexcept nogil # 1BE RES 7,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t RES_1BF(cpu.CPU) noexcept nogil # 1BF RES 7,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1C0(cpu.CPU) noexcept nogil # 1C0 SET 0,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1C1(cpu.CPU) noexcept nogil # 1C1 SET 0,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1C2(cpu.CPU) noexcept nogil # 1C2 SET 0,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1C3(cpu.CPU) noexcept nogil # 1C3 SET 0,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1C4(cpu.CPU) noexcept nogil # 1C4 SET 0,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1C5(cpu.CPU) noexcept nogil # 1C5 SET 0,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1C6(cpu.CPU) noexcept nogil # 1C6 SET 0,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1C7(cpu.CPU) noexcept nogil # 1C7 SET 0,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1C8(cpu.CPU) noexcept nogil # 1C8 SET 1,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1C9(cpu.CPU) noexcept nogil # 1C9 SET 1,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1CA(cpu.CPU) noexcept nogil # 1CA SET 1,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1CB(cpu.CPU) noexcept nogil # 1CB SET 1,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1CC(cpu.CPU) noexcept nogil # 1CC SET 1,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1CD(cpu.CPU) noexcept nogil # 1CD SET 1,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1CE(cpu.CPU) noexcept nogil # 1CE SET 1,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1CF(cpu.CPU) noexcept nogil # 1CF SET 1,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1D0(cpu.CPU) noexcept nogil # 1D0 SET 2,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1D1(cpu.CPU) noexcept nogil # 1D1 SET 2,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1D2(cpu.CPU) noexcept nogil # 1D2 SET 2,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1D3(cpu.CPU) noexcept nogil # 1D3 SET 2,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1D4(cpu.CPU) noexcept nogil # 1D4 SET 2,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1D5(cpu.CPU) noexcept nogil # 1D5 SET 2,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1D6(cpu.CPU) noexcept nogil # 1D6 SET 2,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1D7(cpu.CPU) noexcept nogil # 1D7 SET 2,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1D8(cpu.CPU) noexcept nogil # 1D8 SET 3,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1D9(cpu.CPU) noexcept nogil # 1D9 SET 3,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1DA(cpu.CPU) noexcept nogil # 1DA SET 3,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1DB(cpu.CPU) noexcept nogil # 1DB SET 3,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1DC(cpu.CPU) noexcept nogil # 1DC SET 3,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1DD(cpu.CPU) noexcept nogil # 1DD SET 3,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1DE(cpu.CPU) noexcept nogil # 1DE SET 3,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1DF(cpu.CPU) noexcept nogil # 1DF SET 3,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1E0(cpu.CPU) noexcept nogil # 1E0 SET 4,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1E1(cpu.CPU) noexcept nogil # 1E1 SET 4,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1E2(cpu.CPU) noexcept nogil # 1E2 SET 4,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1E3(cpu.CPU) noexcept nogil # 1E3 SET 4,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1E4(cpu.CPU) noexcept nogil # 1E4 SET 4,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1E5(cpu.CPU) noexcept nogil # 1E5 SET 4,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1E6(cpu.CPU) noexcept nogil # 1E6 SET 4,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1E7(cpu.CPU) noexcept nogil # 1E7 SET 4,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1E8(cpu.CPU) noexcept nogil # 1E8 SET 5,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1E9(cpu.CPU) noexcept nogil # 1E9 SET 5,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1EA(cpu.CPU) noexcept nogil # 1EA SET 5,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1EB(cpu.CPU) noexcept nogil # 1EB SET 5,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1EC(cpu.CPU) noexcept nogil # 1EC SET 5,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1ED(cpu.CPU) noexcept nogil # 1ED SET 5,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1EE(cpu.CPU) noexcept nogil # 1EE SET 5,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1EF(cpu.CPU) noexcept nogil # 1EF SET 5,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1F0(cpu.CPU) noexcept nogil # 1F0 SET 6,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1F1(cpu.CPU) noexcept nogil # 1F1 SET 6,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1F2(cpu.CPU) noexcept nogil # 1F2 SET 6,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1F3(cpu.CPU) noexcept nogil # 1F3 SET 6,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1F4(cpu.CPU) noexcept nogil # 1F4 SET 6,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1F5(cpu.CPU) noexcept nogil # 1F5 SET 6,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1F6(cpu.CPU) noexcept nogil # 1F6 SET 6,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1F7(cpu.CPU) noexcept nogil # 1F7 SET 6,A
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1F8(cpu.CPU) noexcept nogil # 1F8 SET 7,B
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1F9(cpu.CPU) noexcept nogil # 1F9 SET 7,C
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1FA(cpu.CPU) noexcept nogil # 1FA SET 7,D
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1FB(cpu.CPU) noexcept nogil # 1FB SET 7,E
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1FC(cpu.CPU) noexcept nogil # 1FC SET 7,H
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1FD(cpu.CPU) noexcept nogil # 1FD SET 7,L
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1FE(cpu.CPU) noexcept nogil # 1FE SET 7,(HL)
@cython.locals(v=int, flag=uint8_t, t=int)
cdef uint8_t SET_1FF(cpu.CPU) noexcept nogil # 1FF SET 7,A



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

import os.path
from pathlib import Path

import PIL
import pytest

from pyboy import PyBoy, WindowEvent

OVERWRITE_PNGS = False


# https://github.com/aaaaaa123456789/rtc3test
@pytest.mark.skip("RTC is too unstable")
@pytest.mark.parametrize("subtest", [0, 1, 2])
def test_rtc3test(subtest, rtc3test_file):
    pyboy = PyBoy(rtc3test_file, window_type="headless")
    pyboy.set_emulation_speed(0)
    for _ in range(59):
        pyboy.tick()

    for _ in range(25):
        pyboy.tick()

    for n in range(subtest):
        pyboy.send_input(WindowEvent.PRESS_ARROW_DOWN)
        pyboy.tick()
        pyboy.send_input(WindowEvent.RELEASE_ARROW_DOWN)
        pyboy.tick()

    pyboy.send_input(WindowEvent.PRESS_BUTTON_A)
    pyboy.tick()
    pyboy.send_input(WindowEvent.RELEASE_BUTTON_A)
    pyboy.tick()

    while True:
        # Continue until it says "(A) Return"
        if pyboy.botsupport_manager().tilemap_background()[6:14, 17] == [193, 63, 27, 40, 55, 56, 53, 49]:
            break
        pyboy.tick()

    png_path = Path(f"tests/test_results/{rtc3test_file}_{subtest}.png")
    image = pyboy.botsupport_manager().screen().screen_image()
    if OVERWRITE_PNGS:
        png_path.parents[0].mkdir(parents=True, exist_ok=True)
        image.save(png_path)
    else:
        old_image = PIL.Image.open(png_path)
        diff = PIL.ImageChops.difference(image, old_image)
        if diff.getbbox() and not os.environ.get("TEST_CI"):
            image.show()
            old_image.show()
            diff.show()
        assert not diff.getbbox(), f"Images are different! {rtc3test_file}_{subtest}"

    pyboy.stop(save=False)



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#
__pdoc__ = {
    "GameWrapperPokemonGen1.cartridge_title": False,
    "GameWrapperPokemonGen1.post_tick": False,
}

import numpy as np

import pyboy
from pyboy import utils
from pyboy.utils import WindowEvent

from .base_plugin import PyBoyGameWrapper

logger = pyboy.logging.get_logger(__name__)


class GameWrapperPokemonGen1(PyBoyGameWrapper):
    """
    This class wraps Pokemon Red/Blue, and provides basic access for AIs.

    If you call `print` on an instance of this object, it will show an overview of everything this object provides.
    """
    cartridge_title = None

    def __init__(self, *args, **kwargs):
        self.shape = (20, 18)
        super().__init__(*args, game_area_section=(0, 0) + self.shape, game_area_wrap_around=True, **kwargs)
        self.sprite_offset = 0x1000

    def enabled(self):
        return self.pyboy_argv.get("game_wrapper") and ((self.pyboy.cartridge_title() == "POKEMON RED") or
                                                        (self.pyboy.cartridge_title() == "POKEMON BLUE"))

    def post_tick(self):
        self._tile_cache_invalid = True
        self._sprite_cache_invalid = True

        scanline_parameters = self.pyboy.botsupport_manager().screen().tilemap_position_list()
        WX = scanline_parameters[0][2]
        WY = scanline_parameters[0][3]
        self.use_background(WY != 0)

    def _get_screen_background_tilemap(self):
        ### SIMILAR TO CURRENT pyboy.game_wrapper()._game_area_np(), BUT ONLY FOR BACKGROUND TILEMAP, SO NPC ARE SKIPPED
        bsm = self.pyboy.botsupport_manager()
        ((scx, scy), (wx, wy)) = bsm.screen().tilemap_position()
        tilemap = np.array(bsm.tilemap_background()[:, :])
        return np.roll(np.roll(tilemap, -scy // 8, axis=0), -scx // 8, axis=1)[:18, :20]

    def _get_screen_walkable_matrix(self):
        walkable_tiles_indexes = []
        collision_ptr = self.pyboy.get_memory_value(0xD530) + (self.pyboy.get_memory_value(0xD531) << 8)
        tileset_type = self.pyboy.get_memory_value(0xFFD7)
        if tileset_type > 0:
            grass_tile_index = self.pyboy.get_memory_value(0xD535)
            if grass_tile_index != 0xFF:
                walkable_tiles_indexes.append(grass_tile_index + 0x100)
        for i in range(0x180):
            tile_index = self.pyboy.get_memory_value(collision_ptr + i)
            if tile_index == 0xFF:
                break
            else:
                walkable_tiles_indexes.append(tile_index + 0x100)
        screen_tiles = self._get_screen_background_tilemap()
        bottom_left_screen_tiles = screen_tiles[1:1 + screen_tiles.shape[0]:2, ::2]
        walkable_matrix = np.isin(bottom_left_screen_tiles, walkable_tiles_indexes).astype(np.uint8)
        return walkable_matrix

    def game_area_collision(self):
        width = self.game_area_section[2]
        height = self.game_area_section[3]
        game_area = np.ndarray(shape=(height, width), dtype=np.uint32)
        _collision = self._get_screen_walkable_matrix()
        for i in range(height // 2):
            for j in range(width // 2):
                game_area[i * 2][j * 2:j*2 + 2] = _collision[i][j]
                game_area[i*2 + 1][j * 2:j*2 + 2] = _collision[i][j]
        return game_area

    def __repr__(self):
        adjust = 4
        # yapf: disable
        return (
            f"Pokemon Gen 1:\n" +
            "Sprites on screen:\n" +
            "\n".join([str(s) for s in self._sprites_on_screen()]) +
            "\n" +
            "Tiles on screen:\n" +
            " "*5 + "".join([f"{i: <4}" for i in range(10)]) + "\n" +
            "_"*(adjust*20+4) +
            "\n" +
            "\n".join(
                [
                    f"{i: <3}| " + "".join([str(tile).ljust(adjust) for tile in line])
                    for i, line in enumerate(self.game_area())
                ]
            )
        )
        # yapf: enable



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

import pyboy
from pyboy import utils

from .base_mbc import BaseMBC

logger = pyboy.logging.get_logger(__name__)


class MBC3(BaseMBC):
    def setitem(self, address, value):
        if 0x0000 <= address < 0x2000:
            if (value & 0b00001111) == 0b1010:
                self.rambank_enabled = True
            elif value == 0:
                self.rambank_enabled = False
            else:
                # Pan Docs: "Practically any value with 0Ah in the
                # lower 4 bits enables RAM, and any other value
                # disables RAM."
                self.rambank_enabled = False
                # logger.debug("Unexpected command for MBC3: Address: 0x%0.4x, Value: 0x%0.2x", address, value)
        elif 0x2000 <= address < 0x4000:
            value &= 0b01111111
            if value == 0:
                value = 1
            self.rombank_selected = value % self.external_rom_count
        elif 0x4000 <= address < 0x6000:
            self.rambank_selected = value % self.external_ram_count
        elif 0x6000 <= address < 0x8000:
            if self.rtc_enabled:
                self.rtc.writecommand(value)
            # else:
            #     # NOTE: Pokemon Red/Blue will do this, but it can safely be ignored:
            #     # https://github.com/pret/pokered/issues/155
            #     logger.debug("RTC not present. Game tried to issue RTC command: 0x%0.4x, 0x%0.2x", address, value)
        elif 0xA000 <= address < 0xC000:
            if self.rambank_enabled:
                if self.rambank_selected <= 0x03:
                    self.rambanks[self.rambank_selected, address - 0xA000] = value
                elif 0x08 <= self.rambank_selected <= 0x0C:
                    self.rtc.setregister(self.rambank_selected, value)
                # else:
                #     logger.error("Invalid RAM bank selected: 0x%0.2x", self.rambank_selected)
        # else:
        #     logger.error("Invalid writing address: 0x%0.4x", address)



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

cimport pyboy.utils
from pyboy.plugins.base_plugin cimport PyBoyWindowPlugin

import cython

cimport cython
from libc.stdint cimport int16_t, int64_t, uint8_t, uint16_t, uint32_t

from pyboy.logging.logging cimport Logger


cdef Logger logger

cdef int ROWS, COLS

cdef object _sdlcontroller

cpdef list sdl2_event_pump(list) noexcept


cdef class WindowSDL2(PyBoyWindowPlugin):

    cdef int64_t _ftime
    cdef dict _key_down
    cdef dict _key_up
    cdef bint fullscreen

    cdef object _window
    cdef object _sdlrenderer
    cdef object _sdltexturebuffer

    @cython.locals(now=int64_t, delay=int64_t)
    cdef bint frame_limiter(self, int) noexcept



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t

from pyboy.utils cimport IntIOInterface

import cython

from pyboy.logging.logging cimport Logger


cdef Logger logger

cdef class Timer:
    cdef uint64_t DIV, TIMA, TMA, TAC
    cdef uint16_t DIV_counter, TIMA_counter
    cdef uint64_t[4] dividers

    cdef void reset(self) noexcept nogil
    @cython.locals(divider=cython.int)
    cdef bint tick(self, uint64_t) noexcept nogil
    @cython.locals(divider=cython.int, cyclesleft=cython.uint)
    cdef uint64_t cycles_to_interrupt(self) noexcept nogil

    cdef void save_state(self, IntIOInterface) noexcept
    cdef void load_state(self, IntIOInterface, int) noexcept



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

import cython

from cpython.array cimport array
from libc.stdint cimport int16_t, uint8_t, uint16_t, uint32_t, uint64_t

cimport pyboy.utils
from pyboy cimport utils
from pyboy.utils cimport IntIOInterface


cdef uint8_t INTR_VBLANK, INTR_LCDC, INTR_TIMER, INTR_SERIAL, INTR_HIGHTOLOW
cdef uint16_t LCDC, STAT, SCY, SCX, LY, LYC, DMA, BGP, OBP0, OBP1, WY, WX
cdef int ROWS, COLS, TILES, FRAME_CYCLES, VIDEO_RAM, OBJECT_ATTRIBUTE_MEMORY
cdef uint32_t COL0_FLAG
cdef uint8_t BG_PRIORITY_FLAG, CGB_NUM_PALETTES
from pyboy.logging.logging cimport Logger


cdef Logger logger

cdef class LCD:
    cdef bint disable_renderer
    cdef uint8_t[8 * 1024] VRAM0
    cdef uint8_t[0xA0] OAM

    cdef uint8_t SCY
    cdef uint8_t SCX
    cdef uint8_t WY
    cdef uint8_t WX
    cdef uint8_t next_stat_mode
    cdef bint frame_done
    cdef uint8_t max_ly
    cdef uint8_t LY
    cdef uint8_t LYC
    cdef uint64_t clock
    cdef uint64_t clock_target
    cdef LCDCRegister _LCDC
    cdef STATRegister _STAT
    cdef PaletteRegister BGP
    cdef PaletteRegister OBP0
    cdef PaletteRegister OBP1
    cdef Renderer renderer

    @cython.locals(interrupt_flag=uint8_t)
    cdef uint8_t tick(self, int) noexcept nogil
    cdef uint64_t cycles_to_interrupt(self) noexcept nogil

    cdef void set_lcdc(self, uint8_t) noexcept nogil
    cdef uint8_t get_lcdc(self) noexcept nogil
    cdef void set_stat(self, uint8_t) noexcept nogil
    cdef uint8_t get_stat(self) noexcept nogil

    cdef int cycles_to_mode0(self) noexcept nogil

    cdef void save_state(self, IntIOInterface) noexcept
    cdef void load_state(self, IntIOInterface, int) noexcept

    cdef (int, int) getwindowpos(self) noexcept nogil
    cdef (int, int) getviewport(self) noexcept nogil

    # CGB
    cdef bint cgb
    cdef bint double_speed
    cdef uint8_t[8 * 1024] VRAM1
    cdef VBKregister vbk
    cdef PaletteIndexRegister bcps
    cdef PaletteColorRegister bcpd
    cdef PaletteIndexRegister ocps
    cdef PaletteColorRegister ocpd


cdef class PaletteRegister:
    cdef uint8_t value
    cdef uint32_t[4] lookup
    cdef uint32_t[4] palette_mem_rgb

    @cython.locals(x=uint16_t)
    cdef bint set(self, uint64_t) noexcept nogil
    cdef uint8_t get(self) noexcept nogil
    cdef inline uint32_t getcolor(self, uint8_t) noexcept nogil

cdef class STATRegister:
    cdef uint8_t value
    cdef uint8_t _mode
    cdef uint8_t set_mode(self, uint8_t) noexcept nogil
    cdef uint8_t update_LYC(self, uint8_t, uint8_t) noexcept nogil
    cdef void set(self, uint64_t) noexcept nogil

cdef class LCDCRegister:
    cdef uint8_t value

    cdef void set(self, uint64_t) noexcept nogil

    cdef public bint lcd_enable
    cdef public bint windowmap_select
    cdef public bint window_enable
    cdef public bint tiledata_select
    cdef public bint backgroundmap_select
    cdef public bint sprite_height
    cdef public bint sprite_enable
    cdef public bint background_enable
    cdef public bint cgb_master_priority


cdef class Renderer:
    cdef uint8_t[:] _tilecache0_state, _tilecache1_state, _spritecache0_state, _spritecache1_state
    cdef str color_format
    cdef tuple buffer_dims
    cdef bint clearcache
    cdef bint double_speed
    cdef bint cgb

    cdef array _screenbuffer_raw
    cdef object _screenbuffer_ptr
    cdef array _tilecache0_raw, _spritecache0_raw, _spritecache1_raw
    cdef uint32_t[:,:] _screenbuffer
    cdef uint32_t[:,:] _tilecache0, _spritecache0, _spritecache1

    cdef int[10] sprites_to_render
    cdef int ly_window
    cdef void invalidate_tile(self, int, int) noexcept nogil

    cdef int[144][5] _scanlineparameters

    cdef void blank_screen(self, LCD) noexcept nogil

    # CGB
    cdef array _tilecache1_raw
    cdef uint32_t[:,:] _tilecache1

    @cython.locals(
        bx=int,
        by=int,
        wx=int,
        wy=int,
        palette=uint8_t,
        vbank=bint,
        horiflip=bint,
        vertflip=bint,
        bg_priority=bint,
        xx=int,
        yy=int,
        tilecache=uint32_t[:,:],
        bg_priority_apply=uint8_t,
    )
    cdef void scanline(self, LCD, int) noexcept nogil

    @cython.locals(
        spriteheight=int,
        spritecount=int,
        n=int,
        x=int,
        y=int,
        _n=int,
        tileindex=int,
        attributes=int,
        xflip=bint,
        yflip=bint,
        spritepriority=bint,
        palette=uint8_t,
        spritecache=uint32_t[:,:],
        dy=int,
        dx=int,
        yy=int,
        xx=int,
        color_code=uint8_t,
        pixel=uint32_t,
        bgmappriority=bint,
    )
    cdef void scanline_sprites(self, LCD, int, uint32_t[:,:], bint) noexcept nogil
    cdef void sort_sprites(self, int) noexcept nogil

    cdef void clear_cache(self) noexcept nogil
    cdef void clear_tilecache0(self) noexcept nogil
    cdef void clear_tilecache1(self) noexcept nogil # CGB Only
    cdef void clear_spritecache0(self) noexcept nogil
    cdef void clear_spritecache1(self) noexcept nogil
    @cython.locals(
        x=int,
        t=int,
        k=int,
        y=int,
        byte1=uint8_t,
        byte2=uint8_t,
        colorcode=uint32_t,
    )
    cdef void update_tilecache0(self, LCD, int, int) noexcept nogil
    @cython.locals(
        x=int,
        t=int,
        k=int,
        y=int,
        byte1=uint8_t,
        byte2=uint8_t,
        colorcode=uint32_t,
    )
    cdef void update_tilecache1(self, LCD, int, int) noexcept nogil # CGB Only
    @cython.locals(
        x=int,
        t=int,
        k=int,
        y=int,
        byte1=uint8_t,
        byte2=uint8_t,
        colorcode=uint32_t,
    )
    cdef void update_spritecache0(self, LCD, int, int) noexcept nogil
    @cython.locals(
        x=int,
        t=int,
        k=int,
        y=int,
        byte1=uint8_t,
        byte2=uint8_t,
        colorcode=uint32_t,
    )
    cdef void update_spritecache1(self, LCD, int, int) noexcept nogil


    cdef void save_state(self, IntIOInterface) noexcept
    cdef void load_state(self, IntIOInterface, int) noexcept

    @cython.locals(
        tile_num = uint8_t,
        palette = uint8_t,
        vbank = uint8_t,
        horiflip = uint8_t,
        vertflip = uint8_t,
        bg_priority = uint8_t,
    )
    cdef (int, int, int, int, int) _cgb_get_background_map_attributes(self, LCD, int) noexcept nogil

cdef class CGBLCD(LCD):
    pass
    # cdef uint8_t[8 * 1024] VRAM1
    # cdef VBKregister vbk
    # cdef PaletteIndexRegister bcps
    # cdef PaletteColorRegister bcpd
    # cdef PaletteIndexRegister ocps
    # cdef PaletteColorRegister ocpd

cdef class CGBRenderer(Renderer):
    pass

cdef class VBKregister:
    cdef uint8_t active_bank

    cdef void set(self, uint8_t) noexcept nogil
    cdef uint8_t get(self) noexcept nogil

cdef class PaletteIndexRegister:
    cdef uint8_t value
    cdef int auto_inc
    cdef int index
    cdef int hl

    cdef void set(self, uint8_t) noexcept nogil
    cdef uint8_t get(self) noexcept nogil
    cdef int getindex(self) noexcept nogil
    cdef void shouldincrement(self) noexcept nogil

    cdef void save_state(self, IntIOInterface) noexcept
    cdef void load_state(self, IntIOInterface, int) noexcept

cdef class PaletteColorRegister:
    cdef uint16_t[8 * 4] palette_mem
    cdef uint32_t[8 * 4] palette_mem_rgb
    cdef PaletteIndexRegister index_reg

    cdef uint32_t cgb_to_rgb(self, uint16_t, uint8_t) noexcept nogil
    cdef void set(self, uint16_t) noexcept nogil
    cdef uint16_t get(self) noexcept nogil
    cdef inline uint32_t getcolor(self, uint8_t, uint8_t) noexcept nogil

    cdef void save_state(self, IntIOInterface) noexcept
    cdef void load_state(self, IntIOInterface, int) noexcept



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#
"""
The Game Boy has two tile maps, which defines what is rendered on the screen.
"""

import numpy as np
from pyboy.core.lcd import LCDCRegister

from .constants import HIGH_TILEMAP, LCDC_OFFSET, LOW_TILEDATA_NTILES, LOW_TILEMAP
from .tile import Tile


class TileMap:
    def __init__(self, mb, select):
        """
        The Game Boy has two tile maps, which defines what is rendered on the screen. These are also referred to as
        "background" and "window".

        Use `pyboy.botsupport.BotSupportManager.tilemap_background` and
        `pyboy.botsupport.BotSupportManager.tilemap_window` to instantiate this object.

        This object defines `__getitem__`, which means it can be accessed with the square brackets to get a tile
        identifier at a given coordinate.

        Example:
        ```
        >>> tilemap = pyboy.tilemap_window
        >>> tile = tilemap[10,10]
        >>> print(tile)
        34
        >>> print(tilemap[0:10,10])
        [43, 54, 23, 23, 23, 54, 12, 54, 54, 23]
        >>> print(tilemap[0:10,0:4])
        [[43, 54, 23, 23, 23, 54, 12, 54, 54, 23],
         [43, 54, 43, 23, 23, 43, 12, 39, 54, 23],
         [43, 54, 23, 12, 87, 54, 12, 54, 21, 23],
         [43, 54, 23, 43, 23, 87, 12, 50, 54, 72]]
        ```

        Each element in the matrix, is the tile identifier of the tile to be shown on screen for each position. If you
        need the entire 32x32 tile map, you can use the shortcut: `tilemap[:,:]`.
        """
        self.mb = mb
        self._select = select
        self._use_tile_objects = False
        self.refresh_lcdc()

        self.shape = (32, 32)
        """
        Tile maps are always 32x32 tiles.

        Returns
        -------
        (int, int):
            The width and height of the tile map.
        """

    def refresh_lcdc(self):
        """
        The tile data and view that is showed on the background and window respectively can change dynamically. If you
        believe it has changed, you can use this method to update the tilemap from the LCDC register.
        """
        LCDC = LCDCRegister(self.mb.getitem(LCDC_OFFSET))
        if self._select == "WINDOW":
            self.map_offset = HIGH_TILEMAP if LCDC.windowmap_select else LOW_TILEMAP
            self.signed_tile_data = not bool(LCDC.tiledata_select)
        elif self._select == "BACKGROUND":
            self.map_offset = HIGH_TILEMAP if LCDC.backgroundmap_select else LOW_TILEMAP
            self.signed_tile_data = not bool(LCDC.tiledata_select)
        else:
            raise KeyError(f"Invalid tilemap selected: {self._select}")

    def search_for_identifiers(self, identifiers):
        """
        Provided a list of tile identifiers, this function will find all occurrences of these in the tilemap and return
        the coordinates where each identifier is found.

        Example:
        ```
        >>> tilemap = pyboy.tilemap_window
        >>> print(tilemap.search_for_identifiers([43, 123]))
        [[[0,0], [2,4], [8,7]], []]
        ```

        Meaning, that tile identifier `43` is found at the positions: (0,0), (2,4), and (8,7), while tile identifier
        `123`was not found anywhere.

        Args:
            identifiers (list): List of tile identifiers (int)

        Returns
        -------
        list:
            list of matches for every tile identifier in the input
        """
        # TODO: Crude implementation
        tilemap_identifiers = np.asarray(self[:, :], dtype=np.uint32)
        matches = []
        for i in identifiers:
            matches.append([[int(y) for y in x] for x in np.argwhere(tilemap_identifiers == i)])
        return matches

    def _tile_address(self, column, row):
        """
        Returns the memory address in the tilemap for the tile at the given coordinate. The address contains the index
        of tile which will be shown at this position. This should not be confused with the actual tile data of
        `pyboy.botsupport.tile.Tile.data_address`.

        This can be used as an global identifier for the specific location in a tile map.

        Be aware, that the tile index referenced at the memory address might change between calls to
        `pyboy.PyBoy.tick`. And the tile data for the same tile index might also change to display something else
        on the screen.

        The index might also be a signed number. Depending on if it is signed or not, will change where the tile data
        is read from. Use `pyboy.botsupport.tilemap.TileMap.signed_tile_index` to test if the indexes are signed for
        this tile view. You can read how the indexes work in the
        [Pan Docs: VRAM Tile Data](http://bgb.bircd.org/pandocs.htm#vramtiledata).

        Args:
            column (int): Column in this tile map.
            row (int): Row in this tile map.

        Returns
        -------
        int:
            Address in the tile map to read a tile index.
        """

        if not 0 <= column < 32:
            raise IndexError("column is out of bounds. Value of 0 to 31 is allowed")
        if not 0 <= row < 32:
            raise IndexError("row is out of bounds. Value of 0 to 31 is allowed")
        return self.map_offset + 32*row + column

    def tile(self, column, row):
        """
        Provides a `pyboy.botsupport.tile.Tile`-object which allows for easy interpretation of the tile data. The
        object is agnostic to where it was found in the tilemap. I.e. equal `pyboy.botsupport.tile.Tile`-objects might
        be returned from two different coordinates in the tile map if they are shown different places on the screen.

        Args:
            column (int): Column in this tile map.
            row (int): Row in this tile map.

        Returns
        -------
        `pyboy.botsupport.tile.Tile`:
            Tile object corresponding to the tile index at the given coordinate in the
            tile map.
        """
        return Tile(self.mb, self.tile_identifier(column, row))

    def tile_identifier(self, column, row):
        """
        Returns an identifier (integer) of the tile at the given coordinate in the tile map. The identifier can be used
        to quickly recognize what is on the screen through this tile view.

        This identifier unifies the otherwise complicated indexing system on the Game Boy into a single range of
        0-383 (both included).

        You can read how the indexes work in the
        [Pan Docs: VRAM Tile Data](http://bgb.bircd.org/pandocs.htm#vramtiledata).

        Args:
            column (int): Column in this tile map.
            row (int): Row in this tile map.

        Returns
        -------
        int:
            Tile identifier.
        """

        tile = self.mb.getitem(self._tile_address(column, row))
        if self.signed_tile_data:
            return ((tile ^ 0x80) - 128) + LOW_TILEDATA_NTILES
        else:
            return tile

    def __repr__(self):
        adjust = 4
        _use_tile_objects = self._use_tile_objects
        self.use_tile_objects(False)
        # yapf: disable
        return_data = (
            f"Tile Map Address: {self.map_offset:#0{6}x}, " +
            f"Signed Tile Data: {'Yes' if self.signed_tile_data else 'No'}\n" +
            " "*5 + "".join([f"{i: <4}" for i in range(32)]) + "\n" +
            "_"*(adjust*32+2) + "\n" +
            "\n".join(
                [
                    f"{i: <3}| " + "".join([str(tile).ljust(adjust) for tile in line])
                    for i, line in enumerate(self[:, :])
                ]
            )
        )
        # yapf: enable
        self.use_tile_objects(_use_tile_objects)
        return return_data

    def use_tile_objects(self, switch):
        """
        Used to change which object is returned when using the ``__getitem__`` method (i.e. `tilemap[0,0]`).

        Args:
            switch (bool): If True, accesses will return `pyboy.botsupport.tile.Tile`-object. If False, accesses will
                return an `int`.
        """
        self._use_tile_objects = switch

    def __getitem__(self, xy):
        x, y = xy

        if x == slice(None):
            x = slice(0, 32, 1)

        if y == slice(None):
            y = slice(0, 32, 1)

        x_slice = isinstance(x, slice) # Assume slice, otherwise int
        y_slice = isinstance(y, slice) # Assume slice, otherwise int
        assert x_slice or isinstance(x, int)
        assert y_slice or isinstance(y, int)

        if self._use_tile_objects:
            tile_fun = self.tile
        else:
            tile_fun = lambda x, y: self.tile_identifier(x, y)

        if x_slice and y_slice:
            return [[tile_fun(_x, _y) for _x in range(x.stop)[x]] for _y in range(y.stop)[y]]
        elif x_slice:
            return [tile_fun(_x, y) for _x in range(x.stop)[x]]
        elif y_slice:
            return [tile_fun(x, _y) for _y in range(y.stop)[y]]
        else:
            return tile_fun(x, y)



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

import cython

from cpython cimport time
from libc.stdint cimport uint8_t, uint16_t, uint64_t

from pyboy cimport utils
from pyboy.logging.logging cimport Logger
from pyboy.utils cimport IntIOInterface


cdef Logger logger

cdef class RTC:
    cdef str filename
    cdef bint latch_enabled
    cdef cython.double timezero
    cdef uint64_t sec_latch
    cdef uint64_t min_latch
    cdef uint64_t hour_latch
    cdef uint64_t day_latch_low
    cdef uint64_t day_latch_high
    cdef uint64_t day_carry
    cdef uint64_t halt

    cdef void stop(self) noexcept
    cdef void save_state(self, IntIOInterface) noexcept
    cdef void load_state(self, IntIOInterface, int) noexcept
    @cython.locals(days=uint64_t)
    cdef void latch_rtc(self) noexcept nogil
    cdef void writecommand(self, uint8_t) noexcept nogil
    cdef uint8_t getregister(self, uint8_t) noexcept nogil
    @cython.locals(t=cython.double, days=uint64_t)
    cdef void setregister(self, uint8_t, uint8_t) noexcept nogil



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

import os
import platform
import sys

import numpy as np
import pytest
from pyboy import PyBoy, WindowEvent

py_version = platform.python_version()[:3]
is_pypy = platform.python_implementation() == "PyPy"


def test_mario_basics(supermarioland_rom):
    pyboy = PyBoy(supermarioland_rom, window_type="dummy", game_wrapper=True)
    pyboy.set_emulation_speed(0)
    assert pyboy.cartridge_title() == "SUPER MARIOLAN"

    mario = pyboy.game_wrapper()
    mario.start_game(world_level=(1, 1))

    assert mario.score == 0
    assert mario.lives_left == 2
    assert mario.time_left == 400
    assert mario.world == (1, 1)
    assert mario.fitness == 0 # A built-in fitness score for AI development
    pyboy.stop()


def test_mario_advanced(supermarioland_rom):
    pyboy = PyBoy(supermarioland_rom, window_type="dummy", game_wrapper=True)
    pyboy.set_emulation_speed(0)
    assert pyboy.cartridge_title() == "SUPER MARIOLAN"

    mario = pyboy.game_wrapper()
    mario.start_game(world_level=(3, 2))
    lives = 99
    mario.set_lives_left(lives)
    pyboy.tick()

    assert mario.score == 0
    assert mario.lives_left == lives
    assert mario.time_left == 400
    assert mario.world == (3, 2)
    assert mario.fitness == 10000*lives + 6510 # A built-in fitness score for AI development
    pyboy.stop()


def test_mario_game_over(supermarioland_rom):
    pyboy = PyBoy(supermarioland_rom, window_type="dummy", game_wrapper=True)
    pyboy.set_emulation_speed(0)

    mario = pyboy.game_wrapper()
    mario.start_game()
    mario.set_lives_left(0)
    pyboy.send_input(WindowEvent.PRESS_ARROW_RIGHT)
    for _ in range(500): # Enough to game over correctly, and not long enough it'll work without setting the lives
        pyboy.tick()
        if mario.game_over():
            break
    pyboy.stop()


@pytest.mark.skipif(is_pypy, reason="This requires gym, which doesn't work on this platform")
class TestOpenAIGym:
    def test_observation_type_compressed(self, supermarioland_rom):
        pyboy = PyBoy(supermarioland_rom, window_type="dummy", game_wrapper=True)
        pyboy.set_emulation_speed(0)

        env = pyboy.openai_gym(observation_type="compressed")
        if env is None:
            raise Exception("'env' is None. Did you remember to install 'gym'?")
        observation = env.reset()

        expected_observation = np.zeros_like(observation)
        expected_observation[-4:-2, 4:6] = 1 # Mario
        expected_observation[-2:, :] = 10 # Ground
        expected_observation[-4:-2, 1:3] = 14 # Pipe
        expected_observation[9, 5] = 13 # ? Block

        print(observation)
        print(expected_observation)
        assert np.all(observation == expected_observation)

    def test_observation_type_minimal(self, supermarioland_rom):
        pyboy = PyBoy(supermarioland_rom, window_type="dummy", game_wrapper=True)
        pyboy.set_emulation_speed(0)

        env = pyboy.openai_gym(observation_type="minimal")
        if env is None:
            raise Exception("'env' is None. Did you remember to install 'gym'?")
        observation = env.reset()

        expected_observation = np.zeros_like(observation)
        expected_observation[-4:-2, 4:6] = 1 # Mario
        expected_observation[-2:, :] = 3 # Ground
        expected_observation[-4:-2, 1:3] = 3 # Pipe
        expected_observation[9, 5] = 3 # ? Block

        print(observation)
        print(expected_observation)
        assert np.all(observation == expected_observation)

    def test_start_level(self, supermarioland_rom):
        pyboy = PyBoy(supermarioland_rom, window_type="dummy", game_wrapper=True)
        pyboy.set_emulation_speed(0)

        starting_level = (2, 1)
        env = pyboy.openai_gym(observation_type="minimal", action_type="toggle", world_level=starting_level)
        if env is None:
            raise Exception("'env' is None. Did you remember to install 'gym'?")
        observation = env.reset()

        print(env.game_wrapper.world, starting_level)
        assert env.game_wrapper.world == starting_level

        env.game_wrapper.set_lives_left(0)
        env.step(4) # PRESS LEFT

        for _ in range(200):
            env.step(0)
        assert env.game_wrapper.time_left == 399
        assert env.game_wrapper.world == starting_level



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

from libc.stdint cimport uint8_t, uint16_t

from pyboy.logging.logging cimport Logger


cdef Logger logger

cdef class BootROM:
    cdef uint8_t[:] bootrom
    cdef uint8_t getitem(self, uint16_t) noexcept nogil



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

import os
import platform
from pathlib import Path

import pytest

record_gif_py = """
import time
import os
import io
import pickle
import numpy as np
import torch
from pyboy import utils
import sys

from datetime import datetime
from core.gen_algo import get_score, Population, Network
from core.utils import check_needed_turn, do_action, drop_down, \
    do_sideway, do_turn, check_needed_dirs, feature_names
from pyboy import PyBoy, WindowEvent
from multiprocessing import Pool, cpu_count

logger = utils.getLogger("tetris")
logger.setLevel(logging.INFO)

fh = logging.FileHandler('logs.out')
fh.setLevel(logging.INFO)
logger.addHandler(fh)

ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
logger.addHandler(ch)

epochs = 50
population = None
run_per_child = 3
max_fitness = 0
pop_size = 50
max_score = 999999
n_workers = cpu_count()


frames = []


def eval_network(epoch, child_index, child_model, record_to):
    pyboy = PyBoy('tetris_1.1.gb', game_wrapper=True, window_type="headless")
    pyboy.set_emulation_speed(0)
    tetris = pyboy.game_wrapper()
    tetris.start_game()
    pyboy._rendering(False)
    pyboy.send_input(WindowEvent.SCREEN_RECORDING_TOGGLE)

    # Set block animation to fall instantly
    pyboy.set_memory_value(0xff9a, 2)

    run = 0
    scores = []
    levels = []
    lines = []

    while run < run_per_child:
        # Beginning of action
        best_action_score = np.NINF
        best_action = {'Turn': 0, 'Left': 0, 'Right': 0}
        begin_state = io.BytesIO()
        begin_state.seek(0)
        pyboy.save_state(begin_state)
        # Number of lines at the start
        s_lines = tetris.lines

        # Determine how many possible rotations we need to check for the block
        block_tile = pyboy.get_memory_value(0xc203)
        turns_needed = check_needed_turn(block_tile)
        lefts_needed, rights_needed = check_needed_dirs(block_tile)

        # Do middle
        for move_dir in do_action('Middle', pyboy, n_dir=1,
                                  n_turn=turns_needed):
            score = get_score(tetris, child_model, s_lines)
            if score is not None and score >= best_action_score:
                best_action_score = score
                best_action = {'Turn': move_dir['Turn'],
                               'Left': move_dir['Left'],
                               'Right': move_dir['Right']}
            begin_state.seek(0)
            pyboy.load_state(begin_state)

        # Do left
        for move_dir in do_action('Left', pyboy, n_dir=lefts_needed,
                                  n_turn=turns_needed):
            score = get_score(tetris, child_model, s_lines)
            if score is not None and score >= best_action_score:
                best_action_score = score
                best_action = {'Turn': move_dir['Turn'],
                               'Left': move_dir['Left'],
                               'Right': move_dir['Right']}
            begin_state.seek(0)
            pyboy.load_state(begin_state)

        # Do right
        for move_dir in do_action('Right', pyboy, n_dir=rights_needed,
                                  n_turn=turns_needed):
            score = get_score(tetris, child_model, s_lines)
            if score is not None and score >= best_action_score:
                best_action_score = score
                best_action = {'Turn': move_dir['Turn'],
                               'Left': move_dir['Left'],
                               'Right': move_dir['Right']}
            begin_state.seek(0)
            pyboy.load_state(begin_state)

        # Do best action
        for _ in range(best_action['Turn']):
            do_turn(pyboy)
        for _ in range(best_action['Left']):
            do_sideway(pyboy, 'Left')
        for _ in range(best_action['Right']):
            do_sideway(pyboy, 'Right')
        drop_down(pyboy)
        pyboy._rendering(True)
        pyboy.tick()
        frames.append(pyboy.screen_image())
        pyboy._rendering(False)
        if len(frames) >= record_to:
            directory = os.path.join(os.path.curdir, "recordings")
            if not os.path.exists(directory):
                os.makedirs(directory, mode=0o755)
            path = os.path.join(directory, time.strftime(f"{pyboy.cartridge_title()}-%Y.%m.%d-%H.%M.%S.gif"))

            frames[0].save(
                path,
                save_all=True,
                interlace=False,
                loop=0,
                optimize=True,
                append_images=frames[1:],
                duration=int(round(1000 / 30, -1))
            )
            pyboy.stop()
            exit(0)

        # Game over:
        if tetris.game_over() or tetris.score == max_score:
            scores.append(tetris.score)
            levels.append(tetris.level)
            lines.append(tetris.lines)
            if run == run_per_child - 1:
                pyboy.stop()
            else:
                tetris.reset_game()
            run += 1

    child_fitness = np.average(scores)
    logger.info("-" * 20)
    logger.info("Iteration %s - child %s" % (epoch, child_index))
    logger.info("Score: %s, Level: %s, Lines %s" % (scores, levels, lines))
    logger.info("Fitness: %s" % child_fitness)
    logger.info("Output weight:")
    weights = {}
    for i, j in zip(feature_names, child_model.output.weight.data.tolist()[0]):
        weights[i] = np.round(j, 3)
    logger.info(weights)

    return child_fitness


if __name__ == '__main__':
    state_dict = torch.load('models/best.pkl')
    model = Network()
    model.load_state_dict(state_dict)
    eval_network(0, 0, model, 5)


"""


@pytest.mark.skipif(
    os.path.isfile("extras/README/7.gif") or platform.system() == "Windows",
    reason="This test takes too long for regular use"
)
def test_tetris_ai(git_tetris_ai, tetris_rom):
    script_py = "tetris_gif.py"
    with open(Path(git_tetris_ai) / script_py, "w") as f:
        f.write(record_gif_py)

    root_path = Path("../")
    assert os.system(f'rm -rf {Path(git_tetris_ai) / "recordings"}') == 0
    assert os.system(f"cp {tetris_rom} {Path(git_tetris_ai) / 'tetris_1.1.gb'}") == 0
    assert os.system(f'cd {git_tetris_ai} && . {Path(".venv") / "bin" / "activate"} && python {script_py}') == 0
    assert os.system(f'mv {Path(git_tetris_ai) / "recordings" / "TETRIS*"} {Path("extras/README/7.gif")}') == 0



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#
__pdoc__ = {
    "GameWrapperKirbyDreamLand.cartridge_title": False,
    "GameWrapperKirbyDreamLand.post_tick": False,
}

import pyboy
from pyboy import utils
from pyboy.utils import WindowEvent

from .base_plugin import PyBoyGameWrapper

logger = pyboy.logging.get_logger(__name__)


class GameWrapperKirbyDreamLand(PyBoyGameWrapper):
    """
    This class wraps Kirby Dream Land, and provides easy access to score and a "fitness" score for AIs.

    If you call `print` on an instance of this object, it will show an overview of everything this object provides.
    """
    cartridge_title = "KIRBY DREAM LA"

    def __init__(self, *args, **kwargs):
        self.shape = (20, 16)
        """The shape of the game area"""
        self.score = 0
        """The score provided by the game"""
        self.health = 0
        """The health provided by the game"""
        self.lives_left = 0
        """The lives remaining provided by the game"""
        self._game_over = False
        """The game over state"""
        self.fitness = 0
        """
        A built-in fitness scoring. Taking score, health, and lives left into account.

        .. math::
            fitness = score \\cdot health \\cdot lives\\_left
        """
        super().__init__(*args, game_area_section=(0, 0) + self.shape, game_area_wrap_around=True, **kwargs)

    def post_tick(self):
        self._tile_cache_invalid = True
        self._sprite_cache_invalid = True

        self.score = 0
        score_digits = 5
        for n in range(score_digits):
            self.score += self.pyboy.get_memory_value(0xD06F + n) * 10**(score_digits - n)

        # Check if game is over
        prev_health = self.health
        self.health = self.pyboy.get_memory_value(0xD086)
        if self.lives_left == 0:
            if prev_health > 0 and self.health == 0:
                self._game_over = True

        self.lives_left = self.pyboy.get_memory_value(0xD089) - 1

        if self.game_has_started:
            self.fitness = self.score * self.health * self.lives_left

    def start_game(self, timer_div=None):
        """
        Call this function right after initializing PyBoy. This will navigate through menus to start the game at the
        first playable state.

        The state of the emulator is saved, and using `reset_game`, you can get back to this point of the game
        instantly.

        Kwargs:
            timer_div (int): Replace timer's DIV register with this value. Use `None` to randomize.
        """
        PyBoyGameWrapper.start_game(self, timer_div=timer_div)

        # Boot screen
        while True:
            self.pyboy.tick()
            self.tilemap_background.refresh_lcdc()
            if self.tilemap_background[0:3, 16] == [231, 224, 235]: # 'HAL' on the first screen
                break

        # Wait for transition to finish (start screen)
        for _ in range(25):
            self.pyboy.tick()

        self.pyboy.send_input(WindowEvent.PRESS_BUTTON_START)
        self.pyboy.tick()
        self.pyboy.send_input(WindowEvent.RELEASE_BUTTON_START)

        # Wait for transition to finish (exit start screen, enter level intro screen)
        for _ in range(60):
            self.pyboy.tick()

        # Skip level intro
        self.pyboy.send_input(WindowEvent.PRESS_BUTTON_START)
        self.pyboy.tick()
        self.pyboy.send_input(WindowEvent.RELEASE_BUTTON_START)

        # Wait for transition to finish (exit level intro screen, enter game)
        for _ in range(60):
            self.pyboy.tick()

        self.game_has_started = True

        self.saved_state.seek(0)
        self.pyboy.save_state(self.saved_state)

        self._set_timer_div(timer_div)

    def reset_game(self, timer_div=None):
        """
        After calling `start_game`, you can call this method at any time to reset the game.

        Kwargs:
            timer_div (int): Replace timer's DIV register with this value. Use `None` to randomize.
        """
        PyBoyGameWrapper.reset_game(self, timer_div=timer_div)

        self._set_timer_div(timer_div)

    def game_area(self):
        """
        Use this method to get a matrix of the "game area" of the screen.

        ```text
              0   1   2   3   4   5   6   7   8   9
          ____________________________________________________________________________________
          0  | 383 383 383 301 383 383 383 297 383 383 383 301 383 383 383 297 383 383 383 293
          1  | 383 383 383 383 300 294 295 296 383 383 383 383 300 294 295 296 383 383 299 383
          2  | 311 318 319 320 383 383 383 383 383 383 383 383 383 383 383 383 383 301 383 383
          3  | 383 383 383 321 322 383 383 383 383 383 383 383 383 383 383 383 383 383 300 294
          4  | 383 383 383 383 323 290 291 383 383 383 313 312 311 318 319 320 383 290 291 383
          5  | 383 383 383 383 324 383 383 383 383 315 314 383 383 383 383 321 322 383 383 383
          6  | 383 383 383 383 324 293 292 383 383 316 383 383 383 383 383 383 323 383 383 383
          7  | 383 383 383 383 324 383 383 298 383 317 383 383 383 383 383 383 324 383 383 383
          8  | 319 320 383 383 324 383 383 297 383 317 383 383 383 152 140 383 324 383 383 307
          9  | 383 321 322 383 324 294 295 296 383 325 383 383 383 383 383 383 326 272 274 309
          10 | 383 383 323 383 326 383 383 383 2   18  383 330 331 331 331 331 331 331 331 331
          11 | 274 383 324 272 274 272 274 272 274 272 274 334 328 328 328 328 328 328 328 328
          12 | 331 331 331 331 331 331 331 331 331 331 331 328 328 328 328 328 328 328 328 328
          13 | 328 328 328 277 278 328 328 328 328 328 328 328 328 277 278 328 328 277 278 277
          14 | 328 277 278 279 281 277 278 328 328 277 278 277 278 279 281 277 278 279 281 279
          15 | 278 279 281 280 282 279 281 277 278 279 281 279 281 280 282 279 281 280 282 280
        ```

        Returns
        -------
        memoryview:
            Simplified 2-dimensional memoryview of the screen
        """
        return PyBoyGameWrapper.game_area(self)

    def game_over(self):
        return self._game_over

    def __repr__(self):
        adjust = 4
        # yapf: disable
        return (
            f"Kirby Dream Land:\n" +
            f"Score: {self.score}\n" +
            f"Health: {self.health}\n" +
            f"Lives left: {self.lives_left}\n" +
            f"Fitness: {self.fitness}\n" +
            "Sprites on screen:\n" +
            "\n".join([str(s) for s in self._sprites_on_screen()]) +
            "\n" +
            "Tiles on screen:\n" +
            " "*5 + "".join([f"{i: <4}" for i in range(10)]) + "\n" +
            "_"*(adjust*20+4) +
            "\n" +
            "\n".join(
                [
                    f"{i: <3}| " + "".join([str(tile).ljust(adjust) for tile in line])
                    for i, line in enumerate(self.game_area())
                ]
            )
        )
        # yapf: enable



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

import numpy as np

import pyboy
from pyboy import utils
from pyboy.plugins.base_plugin import PyBoyWindowPlugin
from pyboy.utils import WindowEvent

logger = pyboy.logging.get_logger(__name__)

try:
    import OpenGL.GLUT.freeglut
    from OpenGL.GL import (GL_COLOR_BUFFER_BIT, GL_DEPTH_BUFFER_BIT, GL_RGBA, GL_UNSIGNED_INT_8_8_8_8, glClear,
                           glDrawPixels, glFlush, glPixelZoom)
    from OpenGL.GLUT import (GLUT_KEY_DOWN, GLUT_KEY_LEFT, GLUT_KEY_RIGHT, GLUT_KEY_UP, GLUT_RGBA, GLUT_SINGLE,
                             glutCreateWindow, glutDestroyWindow, glutDisplayFunc, glutGetWindow, glutInit,
                             glutInitDisplayMode, glutInitWindowSize, glutKeyboardFunc, glutKeyboardUpFunc,
                             glutReshapeFunc, glutSetWindowTitle, glutSpecialFunc, glutSpecialUpFunc)
    opengl_enabled = True
except (ImportError, AttributeError):
    opengl_enabled = False

ROWS, COLS = 144, 160


class WindowOpenGL(PyBoyWindowPlugin):
    def __init__(self, pyboy, mb, pyboy_argv):
        super().__init__(pyboy, mb, pyboy_argv)

        if not self.enabled():
            return

        if not glutInit():
            raise Exception("OpenGL couldn't initialize!")
        glutInitDisplayMode(GLUT_SINGLE | GLUT_RGBA)
        glutInitWindowSize(*self._scaledresolution)
        glutCreateWindow("PyBoy")
        glutKeyboardFunc(self._key)
        glutKeyboardUpFunc(self._keyUp)
        glutSpecialFunc(self._spec)
        glutSpecialUpFunc(self._specUp)
        self.events = []

        glPixelZoom(self.scale, self.scale)
        glutReshapeFunc(self._glreshape)
        glutDisplayFunc(self._gldraw)
        logger.warning("OpenGL implementation is incomplete. To limit the frame-rate, set your monitor to 60hz.")

    # Cython does not cooperate with lambdas
    def _key(self, c, x, y):
        self._glkeyboard(c.decode("ascii"), x, y, False)

    def _keyUp(self, c, x, y):
        self._glkeyboard(c.decode("ascii"), x, y, True)

    def _spec(self, c, x, y):
        self._glkeyboardspecial(c, x, y, False)

    def _specUp(self, c, x, y):
        self._glkeyboardspecial(c, x, y, True)

    def set_title(self, title):
        glutSetWindowTitle(title)

    def handle_events(self, events):
        events += self.events
        self.events = []
        return events

    def _glkeyboardspecial(self, c, x, y, up):
        if up:
            if c == GLUT_KEY_UP:
                self.events.append(WindowEvent(WindowEvent.RELEASE_ARROW_UP))
            if c == GLUT_KEY_DOWN:
                self.events.append(WindowEvent(WindowEvent.RELEASE_ARROW_DOWN))
            if c == GLUT_KEY_LEFT:
                self.events.append(WindowEvent(WindowEvent.RELEASE_ARROW_LEFT))
            if c == GLUT_KEY_RIGHT:
                self.events.append(WindowEvent(WindowEvent.RELEASE_ARROW_RIGHT))
        else:
            if c == GLUT_KEY_UP:
                self.events.append(WindowEvent(WindowEvent.PRESS_ARROW_UP))
            if c == GLUT_KEY_DOWN:
                self.events.append(WindowEvent(WindowEvent.PRESS_ARROW_DOWN))
            if c == GLUT_KEY_LEFT:
                self.events.append(WindowEvent(WindowEvent.PRESS_ARROW_LEFT))
            if c == GLUT_KEY_RIGHT:
                self.events.append(WindowEvent(WindowEvent.PRESS_ARROW_RIGHT))

    def _glkeyboard(self, c, x, y, up):
        if up:
            if c == "a":
                self.events.append(WindowEvent(WindowEvent.RELEASE_BUTTON_A))
            elif c == "s":
                self.events.append(WindowEvent(WindowEvent.RELEASE_BUTTON_B))
            elif c == "z":
                self.events.append(WindowEvent(WindowEvent.STATE_SAVE))
            elif c == "x":
                self.events.append(WindowEvent(WindowEvent.STATE_LOAD))
            elif c == " ":
                self.events.append(WindowEvent(WindowEvent.RELEASE_SPEED_UP))
            elif c == chr(8):
                self.events.append(WindowEvent(WindowEvent.RELEASE_BUTTON_SELECT))
            elif c == chr(13):
                self.events.append(WindowEvent(WindowEvent.RELEASE_BUTTON_START))
            elif c == "o":
                self.events.append(WindowEvent(WindowEvent.SCREENSHOT_RECORD))
        else:
            if c == "a":
                self.events.append(WindowEvent(WindowEvent.PRESS_BUTTON_A))
            elif c == "s":
                self.events.append(WindowEvent(WindowEvent.PRESS_BUTTON_B))
            elif c == chr(27):
                self.events.append(WindowEvent(WindowEvent.QUIT))
            elif c == " ":
                self.events.append(WindowEvent(WindowEvent.PRESS_SPEED_UP))
            elif c == "i":
                self.events.append(WindowEvent(WindowEvent.SCREEN_RECORDING_TOGGLE))
            elif c == chr(8):
                self.events.append(WindowEvent(WindowEvent.PRESS_BUTTON_SELECT))
            elif c == chr(13):
                self.events.append(WindowEvent(WindowEvent.PRESS_BUTTON_START))

    def _glreshape(self, width, height):
        scale = max(min(height / ROWS, width / COLS), 1)
        self._scaledresolution = (round(scale * COLS), round(scale * ROWS))
        glPixelZoom(scale, scale)
        # glutReshapeWindow(*self._scaledresolution);

    def _gldraw(self):
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        buf = np.asarray(self.renderer._screenbuffer)[::-1, :]
        glDrawPixels(COLS, ROWS, GL_RGBA, GL_UNSIGNED_INT_8_8_8_8, buf)
        glFlush()

    def enabled(self):
        if self.pyboy_argv.get("window_type") == "OpenGL":
            if opengl_enabled:
                if bool(OpenGL.GLUT.freeglut.glutMainLoopEvent):
                    return True
                else:
                    logger.error("Failed to load \"PyOpenGL\". OpenGL window disabled")
            else:
                logger.error("Missing depencency \"PyOpenGL\". OpenGL window disabled")
        return False

    def post_tick(self):
        self._gldraw()
        OpenGL.GLUT.freeglut.glutMainLoopEvent()

    def stop(self):
        glutDestroyWindow(glutGetWindow())
        for _ in range(10): # At least 2 to close
            OpenGL.GLUT.freeglut.glutMainLoopEvent()



#
# License: See LICENSE.md file
# GitHub: https://github.com/Baekalfen/PyBoy
#

__pdoc__ = {
    "PyBoyPlugin": False,
    "PyBoyWindowPlugin": False,
    "PyBoyGameWrapper.post_tick": False,
    "PyBoyGameWrapper.enabled": False,
    "PyBoyGameWrapper.argv": False,
}

import io
import random
from array import array

import numpy as np

import pyboy
from pyboy.botsupport.sprite import Sprite

logger = pyboy.logging.get_logger(__name__)

try:
    from cython import compiled
    cythonmode = compiled
except ImportError:
    cythonmode = False

ROWS, COLS = 144, 160


class PyBoyPlugin:
    argv = []

    def __init__(self, pyboy, mb, pyboy_argv):
        if not cythonmode:
            self.pyboy = pyboy
            self.mb = mb
            self.pyboy_argv = pyboy_argv

    def __cinit__(self, pyboy, mb, pyboy_argv, *args, **kwargs):
        self.pyboy = pyboy
        self.mb = mb
        self.pyboy_argv = pyboy_argv

    def handle_events(self, events):
        return events

    def post_tick(self):
        pass

    def window_title(self):
        return ""

    def stop(self):
        pass

    def enabled(self):
        return True


class PyBoyWindowPlugin(PyBoyPlugin):
    def __init__(self, pyboy, mb, pyboy_argv, *args, **kwargs):
        super().__init__(pyboy, mb, pyboy_argv, *args, **kwargs)

        if not self.enabled():
            return

        scale = pyboy_argv.get("scale")
        self.scale = scale
        logger.debug("%s initialization" % self.__class__.__name__)

        self._scaledresolution = (scale * COLS, scale * ROWS)
        logger.debug("Scale: x%d (%d, %d)", self.scale, self._scaledresolution[0], self._scaledresolution[1])

        self.enable_title = True
        if not cythonmode:
            self.renderer = mb.lcd.renderer

    def __cinit__(self, *args, **kwargs):
        self.renderer = self.mb.lcd.renderer

    def frame_limiter(self, speed):
        return False

    def set_title(self, title):
        pass


class PyBoyGameWrapper(PyBoyPlugin):
    """
    This is the base-class for the game-wrappers. It provides some generic game-wrapping functionality, like `game_area`
    , which shows both sprites and tiles on the screen as a simple matrix.
    """

    argv = [("--game-wrapper", {"action": "store_true", "help": "Enable game wrapper for the current game"})]

    def __init__(self, *args, game_area_section=(0, 0, 32, 32), game_area_wrap_around=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.tilemap_background = self.pyboy.botsupport_manager().tilemap_background()
        self.tilemap_window = self.pyboy.botsupport_manager().tilemap_window()
        self.tilemap_use_background = True
        self.sprite_offset = 0
        self.game_has_started = False
        self._tile_cache_invalid = True
        self._sprite_cache_invalid = True

        self.game_area_section = game_area_section
        self.game_area_wrap_around = game_area_wrap_around
        width = self.game_area_section[2] - self.game_area_section[0]
        height = self.game_area_section[3] - self.game_area_section[1]
        self._cached_game_area_tiles_raw = array("B", [0xFF] * (width*height*4))

        self.saved_state = io.BytesIO()

        self._cached_game_area_tiles = memoryview(self._cached_game_area_tiles_raw).cast("I", shape=(width, height))

    def enabled(self):
        return self.pyboy_argv.get("game_wrapper") and self.pyboy.cartridge_title() == self.cartridge_title

    def post_tick(self):
        raise NotImplementedError("post_tick not implemented in game wrapper")

    def _set_timer_div(self, timer_div):
        if timer_div is None:
            self.mb.timer.DIV = random.getrandbits(8)
        else:
            self.mb.timer.DIV = timer_div & 0xFF

    def start_game(self, timer_div=None):
        """
        Call this function right after initializing PyBoy. This will navigate through menus to start the game at the
        first playable state.

        A value can be passed to set the timer's DIV register. Some games depend on it for randomization.

        The state of the emulator is saved, and using `reset_game`, you can get back to this point of the game
        instantly.

        Args:
            timer_div (int): Replace timer's DIV register with this value. Use `None` to randomize.
        """

        if not self.pyboy.frame_count == 0:
            logger.warning("Calling start_game from an already running game. This might not work.")

    def reset_game(self, timer_div=None):
        """
        After calling `start_game`, you can call this method at any time to reset the game.

        Args:
            timer_div (int): Replace timer's DIV register with this value. Use `None` to randomize.
        """

        if self.game_has_started:
            self.saved_state.seek(0)
            self.pyboy.load_state(self.saved_state)
            self.post_tick()
        else:
            logger.error("Tried to reset game, but it hasn't been started yet!")

    def game_over(self):
        """
        After calling `start_game`, you can call this method at any time to know if the game is over.
        """
        raise NotImplementedError("game_over not implemented in game wrapper")

    def _sprites_on_screen(self):
        if self._sprite_cache_invalid:
            self._cached_sprites_on_screen = []
            for s in range(40):
                sprite = Sprite(self.mb, s)
                if sprite.on_screen:
                    self._cached_sprites_on_screen.append(sprite)
            self._sprite_cache_invalid = False
        return self._cached_sprites_on_screen

    def _game_area_tiles(self):
        if self._tile_cache_invalid:
            xx = self.game_area_section[0]
            yy = self.game_area_section[1]
            width = self.game_area_section[2]
            height = self.game_area_section[3]
            scanline_parameters = self.pyboy.botsupport_manager().screen().tilemap_position_list()

            if self.game_area_wrap_around:
                self._cached_game_area_tiles = np.ndarray(shape=(height, width), dtype=np.uint32)
                for y in range(height):
                    SCX = scanline_parameters[(yy+y) * 8][0] // 8
                    SCY = scanline_parameters[(yy+y) * 8][1] // 8
                    for x in range(width):
                        _x = (xx+x+SCX) % 32
                        _y = (yy+y+SCY) % 32
                        if self.tilemap_use_background:
                            self._cached_game_area_tiles[y, x] = self.tilemap_background.tile_identifier(_x, _y)
                        else:
                            self._cached_game_area_tiles[y, x] = self.tilemap_window.tile_identifier(_x, _y)
            else:
                if self.tilemap_use_background:
                    self._cached_game_area_tiles = np.asarray(
                        self.tilemap_background[xx:xx + width, yy:yy + height], dtype=np.uint32
                    )
                else:
                    self._cached_game_area_tiles = np.asarray(
                        self.tilemap_window[xx:xx + width, yy:yy + height], dtype=np.uint32
                    )
            self._tile_cache_invalid = False
        return self._cached_game_area_tiles

    def use_background(self, value):
        self.tilemap_use_background = value

    def game_area(self):
        """
        This method returns a cut-out of the screen as a simplified matrix for use in machine learning applications.

        Returns
        -------
        memoryview:
            Simplified 2-dimensional memoryview of the screen
        """
        tiles_matrix = self._game_area_tiles()
        sprites = self._sprites_on_screen()
        xx = self.game_area_section[0]
        yy = self.game_area_section[1]
        width = self.game_area_section[2]
        height = self.game_area_section[3]
        for s in sprites:
            _x = (s.x // 8) - xx
            _y = (s.y // 8) - yy
            if 0 <= _y < height and 0 <= _x < width:
                tiles_matrix[_y][
                    _x] = s.tile_identifier + self.sprite_offset # Adding offset to try to seperate sprites from tiles
        return tiles_matrix

    def _game_area_np(self, observation_type="tiles"):
        if observation_type == "tiles":
            return np.asarray(self.game_area(), dtype=np.uint16)
        elif observation_type == "compressed":
            try:
                return self.tiles_compressed[np.asarray(self.game_area(), dtype=np.uint16)]
            except AttributeError:
                raise AttributeError(
                    f"Game wrapper miss the attribute tiles_compressed for observation_type : {observation_type}"
                )
        elif observation_type == "minimal":
            try:
                return self.tiles_minimal[np.asarray(self.game_area(), dtype=np.uint16)]
            except AttributeError:
                raise AttributeError(
                    f"Game wrapper miss the attribute tiles_minimal for observation_type : {observation_type}"
                )
        else:
            raise ValueError(f"Invalid observation_type : {observation_type}")

    def _sum_number_on_screen(self, x, y, length, blank_tile_identifier, tile_identifier_offset):
        number = 0
        for i, x in enumerate(self.tilemap_background[x:x + length, y]):
            if x != blank_tile_identifier:
                number += (x+tile_identifier_offset) * (10**(length - 1 - i))
        return number
